import matplotlib
import matplotlib.pyplot as plt
import pandas as pd
from matplotlib.ticker import MultipleLocator
import annotations_trimm
import os
import numpy as np
import types
matplotlib.use("Agg")


def annotations_plot(data: pd.Series, trimm_type: str, variables: dict, size_limit: float,
                     plot_method, data_type: str, title: str) -> None:
    """Creates visualisation basedon plot_method"""
    # plot data with
    new_data = data.reset_index()
    new_data.columns = ['old_index', 'counts']
    new_data['new_index'] = new_data['old_index'].apply(lambda text: annotations_trimm.trimm(text, trimm_type, data_type))
    if data_type == 'clusters':
        if trimm_type != 'repeat type':
            new_data = new_data[new_data['new_index'] != 'not mobile element']

    counts = new_data.groupby(['new_index']).sum('counts').counts

    file_name = data_type + '_' + "_".join(trimm_type.split())
    plot_method(counts, file_name, size_limit, variables, title, data_type)


def clusters_plots(cluster_table: pd.DataFrame, variables: dict, log: list) -> None:
    """Creates clusters plots depend on what type of plot is set"""

    annotations = cluster_table['Automatic_annotation']
    counts = annotations.value_counts()
    annotation_types = "repeat type,class,subclass,order,family".split(',')
    legend_template = "Anotace klastrů dle {} {}"
    legend_elements = "repetitivních elementů,transposibilních elementů," \
                             "transposibilních elementů,transposibilních elementů,transposibilních elementů".split(',')
    legend_annotations = "typu,třídy,podtřídy,řádu,rodiny".split(',')

    for plot_type, plot_method in [("clusters_pp", create_pieplot), ("clusters_bp", create_barplot)]:
        if variables[plot_type].get():
            # types of plots - by repeat type, by TE class + satellite, by TE subclass + satellite, by TE type + satellite

            for annotation_type, element, annotation in zip(annotation_types, legend_elements, legend_annotations):
                annotations_plot(counts, annotation_type, variables, variables['clusters_size_limit'], plot_method,
                                 'clusters', legend_template.format(annotation, element ))
                log.append(f'Vytvořen {plot_type[-2:]} {legend_template.format(annotation, element )}\n')


def reads_plots(variables: dict, annotations: pd.Series, log: list):
    """Creates reads_plot"""

    annotation_types = "class,subclass,order,family,protein domain,protein domain with family".split(',')
    legend_template = "Anotace čtení dle {} transpozibilních elementů"
    legend_annotations = "třídy,podtřídy,řádu,rodiny,proteinové domény, proteinové domény s danou rodinou".split(',')

    for plot_type, plot_method in [("reads_pp", create_pieplot), ("reads_bp", create_barplot)]:
        if variables[plot_type].get():
            # types of output graphs - by class, by subclass, by protein domain, by element, by element domain
            for annotation_type, annotation in zip(annotation_types, legend_annotations):
                annotations_plot(annotations, annotation_type, variables, variables['reads_size_limit'], plot_method,
                                 'reads', legend_template.format(annotation))
                log.append(f'Vytvořen {plot_type[-2:]} s anotovanými čteními dle {annotation} transpozibilních elementů\n')


def tarean_plots(cluster_table: pd.DataFrame, variables: dict, log: list) -> None:
    """Creates all tarean annotations visualisations"""

    annotations = cluster_table['TAREAN_annotation']
    counts = annotations.value_counts()

    for plot_type, plot_method in [("clusters_pp", create_pieplot), ("clusters_bp", create_barplot)]:
        if variables[plot_type].get():
            annotations_plot(counts, 'all', variables, variables['tarean_size_limit'], plot_method,
                             'tarean', 'TAREAN anotace')
            log.append(f'Vytvořen {plot_type[-2:]} s TAREAN anotacemi klastrů\n')

def annotated_clusters_visualization(cluster_table: pd.DataFrame, variables: dict, log: list) -> None:
    """ Function for visualisation of annotated clusters, there are two types of annotation in CLUSTER_TABLE.csv -
    TE annotation (column Automatic_annotation) and TAREAN annotation (column TAREAN_annotation)"""

    # check what plots functions are set and run them
    clusters_plots(cluster_table, variables, log)

    tarean_plots(cluster_table, variables, log)


def extract_reads_files(cl_count: int, cl_dir: str, log: list) -> pd.Series:
    """Extract data from files containing reads information, files are stored in folder specific for cluster """
    annotations = pd.Series(dtype=np.uint32)

    for cl_id in range(1, cl_count + 1):
        cl_id = str(cl_id)
        actual_cl_dir = "dir_CL" + (4 - len(cl_id)) * "0" + cl_id + '/'
        data = pd.read_csv(cl_dir + actual_cl_dir + 'protein_database_annotation.csv', sep='\t', header=0, index_col=0)
        log.append(f"Úspěšně načten soubor {actual_cl_dir}protein_database_annotation.csv\n")
        if not data.empty:
            log.append(f"Soubor obsahuje anotované proteiny\n")

            trimmed_ids = data['db_id'].apply(lambda d: d[d.find('#') + 1:])
            annotations = annotations.add(trimmed_ids.value_counts(), fill_value=0)

    return annotations


def annotated_reads_visualization(cl_count, variables, log):
    cl_dir = variables['input'].get() + 'seqclust/clustering/clusters/'

    annotations = extract_reads_files(cl_count, cl_dir, log)


    reads_plots(variables, annotations, log)

def clusters_structure_visualization(data, title, variables, log):
    sample1 = variables['first_sample'].get()
    sample2 = variables['second_sample'].get()
    sums = data[sample2] + data[sample1]
    means = sums / 2
    ratios = data['00']

    bb_enriched_clusters = data[ratios / sums <= 1 - variables['b_clusters_limit'].get()]

    if variables['cluster_structure'].get():

        fig = plt.figure(figsize=(12, 7), )
        ax1 = fig.add_subplot(111, )

        l1 = ax1.plot(data.index, sums)
        l2 = ax1.plot(data.index, means)
        l3 = ax1.plot(data.index, ratios)

        ax1.set_ylabel('Počet čtení')
        ax1.set_xlabel('ID klastru')
        ax1.set_title(title)
        xticks = [data.index[i] for i in range(0, len(data) + 1, 10)]
        ax1.set_xticks(xticks)
        ax1.xaxis.set_minor_locator(MultipleLocator(1))
        ax1.margins(x=0, y=0)
        ax1.legend(('Velikost klastru', 'Polovina velikosti klastru', 'Počet 00 čtení'))
        log.append("Vytvořen graf popisující strukturu klastrů\n")

        for cl_id in bb_enriched_clusters.index:
            ax1.plot(cl_id, ratios[cl_id], 'r*')

        fig.tight_layout()

        output_dest = variables['output'].get() + 'cluster_overview'

        data.to_csv(output_dest + '.csv', sep='\t')
        plt.savefig(output_dest + '.png')

        plt.close(fig)

    return bb_enriched_clusters


def create_barplot(data, file_name, size_limit, variables, title, data_type):
    # from data of type value_count Series creates a barplot with title and size_limit to a png file_name

    filter_value = round(data.sum() * size_limit.get())
    fig = plt.figure(figsize=(15, 8))

    if data_type == 'reads':
        header = 'Počet čtení'
        xlabel = "Počet čtení"
        output_dest = variables['output'].get() + 'bar_plots/reads/'
    else:
        header = 'Počet klastrů'
        xlabel = "Počet klastrů"
        if data_type == 'clusters':
            output_dest = variables['output'].get() + 'bar_plots/clusters/'
        elif data_type == 'tarean':
            output_dest = variables['output'].get() + 'bar_plots/tarean/'

    filtered_data = data[data > filter_value]
    filtered_data.sort_values(inplace=True)
    colors = plt.get_cmap("viridis")

    ax = filtered_data.plot(kind='barh', position=0.5,
                            color=[colors(i / len(filtered_data)) for i in range(len(filtered_data))])
    ax.set_ylabel('Anotace')
    ax.set_xlabel(xlabel)
    ax.tick_params(axis='y', labelsize=10)
    ax.set_title(title)

    fig.tight_layout()

    try:
        os.makedirs(output_dest)
    except FileExistsError:
        pass

    data.to_csv(output_dest + file_name + '.csv', sep='\t',
                header=[header], index_label='Anotace', encoding='utf-8-sig')
    plt.savefig(output_dest + file_name)

    plt.close(fig)


def create_pieplot(data, file_name, size_limit, variables, title, data_type):
    filter_value = round(data.sum() * size_limit.get())
    fig = plt.figure(figsize=(15, 15))
    filtered_data = data[data > filter_value]
    filtered_data.sort_values(inplace=True)

    ax = fig.add_subplot(111)
    colors = plt.get_cmap("viridis")
    ax.set_prop_cycle("color", [colors(i / len(filtered_data)) for i in range(len(filtered_data))])
    ax.pie(filtered_data.values, autopct=lambda p: round(filtered_data.sum() * p / 100))

    ax.legend(labels=filtered_data.index, title='Anotace', loc='center left', bbox_to_anchor=(1, 0, 0.5, 1))
    ax.axis('equal')
    ax.set_title(title)
    fig.tight_layout()

    if data_type == 'reads':
        header = 'Počet čtení'
        output_dest = variables['output'].get() + 'pie_plots/reads/'
    else:
        header = 'Počet klastrů'
        if data_type == 'clusters':
            output_dest = variables['output'].get() + 'pie_plots/clusters/'
        elif data_type == 'tarean':
            output_dest = variables['output'].get() + 'pie_plots/tarean/'

    try:
        os.makedirs(output_dest)
    except FileExistsError:
        pass

    data.to_csv(output_dest + file_name + '.csv', sep='\t',
                header=[header], index_label='Anotace', encoding='utf-8-sig')
    plt.savefig(output_dest + file_name)

    plt.close(fig)
